/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.1 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <system_settings.h>
#include <Elementary.h>
#include "$(appName).h"
#include "view.h"
#include "data.h"

static void _display_changed_cb(void);
static const char *_get_display_string_cb(void);
static bool _key_pressed_cb(key_type key);
static void _set_region_format(void);

/**
 * @brief Hook to take necessary actions before main event loop starts.
 * Initialize UI resources and application's data.
 * If this function returns true, the main loop of application starts.
 * If this function returns false, the application is terminated.
 */
static bool app_create(void *user_data)
{
	data_initialize();
	data_set_display_change_cb(_display_changed_cb);

	view_set_callbacks(_get_display_string_cb, _key_pressed_cb);

	if (!view_create(NULL))
		return false;

	_set_region_format();	/* This is called to set valid decimal character on the keypad. */

	return true;
}

/**
 * @brief This callback function is called when another application
 * sends a launch request to the application.
 */
static void app_control(app_control_h app_control, void *user_data)
{
	/* Handle the launch request. */
}

/**
 * @brief This callback function is called each time
 * the application is completely obscured by another application
 * and becomes invisible to the user.
 */
static void app_pause(void *user_data)
{
	/* Take necessary actions when application becomes invisible. */
}

/**
 * @brief This callback function is called each time
 * the application becomes visible to the user.
 */
static void app_resume(void *user_data)
{
	/* Take necessary actions when application becomes visible. */
}

/**
 * @brief This callback function is called once after the main loop of the application exits.
 */
static void app_terminate(void *user_data)
{
	view_destroy();
}

/**
 * @brief This function will be called when the language is changed.
 */
static void ui_app_lang_changed(app_event_info_h event_info, void *user_data)
{
	/* APP_EVENT_LANGUAGE_CHANGED */
	char *locale = NULL;

	system_settings_get_value_string(SYSTEM_SETTINGS_KEY_LOCALE_LANGUAGE, &locale);

	if (locale != NULL) {
		elm_language_set(locale);
		free(locale);
	}
	return;
}

/**
 * @brief This function will be called when the region format is changed.
 */
static void ui_app_region_changed(app_event_info_h event_info, void *user_data)
{
	/* APP_EVENT_REGION_FORMAT_CHANGED */
	data_update_region_format();

	_set_region_format();
}

/**
 * @brief This function will be called when the device orientation is changed.
 */
static void ui_app_orientation_changed(app_event_info_h event_info, void *user_data)
{
	/* APP_EVENT_DEVICE_ORIENTATION_CHANGED */
	app_device_orientation_e orientation;
	int ret;

	ret = app_event_get_device_orientation(event_info, &orientation);
	if (ret != APP_ERROR_NONE) {
		dlog_print(DLOG_ERROR, LOG_TAG, "app_event_get_device_orientation() failed. Err = %d", ret);
		return;
	}

	view_change_orientation((int)orientation);
}

/**
 * @brief Main function of the application.
 */
int main(int argc, char *argv[])
{
	int ret;

	ui_app_lifecycle_callback_s event_callback = {0, };
	app_event_handler_h handlers[5] = {NULL, };

	event_callback.create = app_create;
	event_callback.terminate = app_terminate;
	event_callback.pause = app_pause;
	event_callback.resume = app_resume;
	event_callback.app_control = app_control;

	/*
	 * If you want to handle more events,
	 * please check the application lifecycle guide.
	 */
	ui_app_add_event_handler(&handlers[APP_EVENT_LANGUAGE_CHANGED], APP_EVENT_LANGUAGE_CHANGED, ui_app_lang_changed, NULL);
	ui_app_add_event_handler(&handlers[APP_EVENT_REGION_FORMAT_CHANGED], APP_EVENT_REGION_FORMAT_CHANGED, ui_app_region_changed, NULL);
	ui_app_add_event_handler(&handlers[APP_EVENT_DEVICE_ORIENTATION_CHANGED], APP_EVENT_DEVICE_ORIENTATION_CHANGED, ui_app_orientation_changed, NULL);

	ret = ui_app_main(argc, argv, &event_callback, NULL);
	if (ret != APP_ERROR_NONE)
		dlog_print(DLOG_ERROR, LOG_TAG, "ui_app_main() failed. err = %d", ret);

	return ret;
}

/**
 * @brief Internal callback function handling model display change event.
 * @param[in] data The user data passed to the callback attachment function.
 */
static void _display_changed_cb(void)
{
	view_update_display();
}

/**
 * @brief Internal callback function which obtains the calculator's output
 * which will be set to the display.
 * @return This function returns the calculators output in for of a text string.
 * On failure NULL is returned.
 */
static const char *_get_display_string_cb(void)
{
	return data_get_display_str();
}

/**
 * @brief Internal callback function which handles the keypad key press.
 * @return This function returns 'true' if the maximum number of digits
 * is not reached yet, otherwise 'false' is returned.
 */
static bool _key_pressed_cb(key_type key)
{
	data_set_key_press(key);

	return (data_get_last_error() != RES_MAX_DIGITS_REACHED);
}

/**
 * @brief Internal function which gets the valid decimal character according
 * to the current regional format setting and sets it to the view module.
 */
static void _set_region_format(void)
{
	const char *format = data_get_dec_point_str();

	view_change_region_format(format);
}
