/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.1 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <Elementary.h>
#include <efl_extension.h>
#include "view.h"
#include "$(appName).h"

#define WIDGETS_COUNT 20
#define BUTTONS_COUNT WIDGETS_COUNT - 1
#define CIRCLE_QUARTER_DEG 90
#define CIRCLE_HALF_QUARTER_DEG CIRCLE_QUARTER_DEG / 2

#define VIEW_BORDER_WIDTH 10 /* Border around all the grid with widgets */
#define WIDGET_PADDING 3 /* Padding around all widgets (adds extra pixels to the border) */
#define DISP_NUMBER_RIGHT_PADDING 20 /* Padding between the right edge of the display text and grid cell virtual edje */
#define DISP_NUMBER_TEXT_SIZE 56

#define PORT_DISP_HEIGHT 4 /* Height of the button in the logical grid units */
#define PORT_DISP_GAP 6 /* Space between display */
#define PORT_BTN1_HEIGHT 4 /* Height of the first row of the buttons */
#define PORT_AUX_GAP 1 /* Space between first and second row of the buttons */
#define PORT_BTN2_HEIGHT 4 /* Height of the other rows of the buttons */

/* Y offsets of the widget rows for portrait orientation */
#define PORT_OFFS0 0
#define PORT_OFFS1 PORT_OFFS0 + PORT_DISP_HEIGHT + PORT_DISP_GAP
#define PORT_OFFS2 PORT_OFFS1 + PORT_BTN1_HEIGHT + PORT_AUX_GAP
#define PORT_OFFS3 PORT_OFFS2 + PORT_BTN2_HEIGHT
#define PORT_OFFS4 PORT_OFFS3 + PORT_BTN2_HEIGHT
#define PORT_OFFS5 PORT_OFFS4 + PORT_BTN2_HEIGHT
#define PORT_OFFS6 PORT_OFFS5 + PORT_BTN2_HEIGHT

/* Size of the grid in logical units for portrait orientation */
#define PORT_GRID_W 4
#define PORT_GRID_H PORT_OFFS6 + PORT_BTN2_HEIGHT

/* Landscape parameters */
#define LAND_DISP_HEIGHT 4
#define LAND_DISP_GAP 1
#define LAND_BTN_HEIGHT 4
#define LAND_OFFS0 0
#define LAND_OFFS1 LAND_OFFS0 + LAND_DISP_HEIGHT + LAND_DISP_GAP
#define LAND_OFFS2 LAND_OFFS1 + LAND_BTN_HEIGHT
#define LAND_OFFS3 LAND_OFFS2 + LAND_BTN_HEIGHT
#define LAND_GRID_W 7
#define LAND_GRID_H LAND_OFFS3 + LAND_BTN_HEIGHT

/* Internal data structure of the main_view object */
static struct view_info {
	Evas_Object *win;
	Evas_Object *conform;
	Evas_Object *navi;
	Evas_Object *grid;
	Evas_Object *display;
	Evas_Object *button_decimal;
	Evas_Object *keypad[WIDGETS_COUNT];
	get_display_string_cb get_display_string_func;
	key_pressed_cb key_pressed_func;
} s_info = {
	.win = NULL,
	.conform = NULL,
	.navi = NULL,
	.grid = NULL,
	.display = NULL,
	.button_decimal = NULL,
	.keypad = {NULL,},
	.get_display_string_func = NULL,
	.key_pressed_func = NULL,
};

typedef struct {
	key_type key; /* key type of the button or KEY_NONE for label. */
	const char *text; /* text label for the button or NULL for label. */
} widget_params_t;

typedef struct {
	int x;
	int y;
	int w;
	int h;
} widget_pos_t;

static const widget_params_t widget_params[] = {
	{KEY_NONE,   NULL},
	{KEY_RESULT, "="},
	{KEY_SIGN,   "+/-"},
	{KEY_ERASE,  "<font_size=48>\xE2\x86\x90</font_size>"},
	{KEY_RESET,  "<b>C</b>"},
	{KEY_0,      "0"},
	{KEY_1,      "1"},
	{KEY_2,      "2"},
	{KEY_3,      "3"},
	{KEY_4,      "4"},
	{KEY_5,      "5"},
	{KEY_6,      "6"},
	{KEY_7,      "7"},
	{KEY_8,      "8"},
	{KEY_9,      "9"},
	{KEY_DEC,    NULL},
	{KEY_ADD,    "+"},
	{KEY_SUB,    "-"},
	{KEY_MUL,    "*"},
	{KEY_DIV,    "/"}
};

static const widget_pos_t portrait_wpos[] = {
	{0, PORT_OFFS0, 4, PORT_DISP_HEIGHT},
	{3, PORT_OFFS5, 1, PORT_BTN2_HEIGHT * 2},
	{0, PORT_OFFS2, 1, PORT_BTN2_HEIGHT},
	{2, PORT_OFFS1, 2, PORT_BTN1_HEIGHT},
	{0, PORT_OFFS1, 1, PORT_BTN1_HEIGHT},
	{0, PORT_OFFS6, 2, PORT_BTN2_HEIGHT},
	{0, PORT_OFFS5, 1, PORT_BTN2_HEIGHT},
	{1, PORT_OFFS5, 1, PORT_BTN2_HEIGHT},
	{2, PORT_OFFS5, 1, PORT_BTN2_HEIGHT},
	{0, PORT_OFFS4, 1, PORT_BTN2_HEIGHT},
	{1, PORT_OFFS4, 1, PORT_BTN2_HEIGHT},
	{2, PORT_OFFS4, 1, PORT_BTN2_HEIGHT},
	{0, PORT_OFFS3, 1, PORT_BTN2_HEIGHT},
	{1, PORT_OFFS3, 1, PORT_BTN2_HEIGHT},
	{2, PORT_OFFS3, 1, PORT_BTN2_HEIGHT},
	{2, PORT_OFFS6, 1, PORT_BTN2_HEIGHT},
	{3, PORT_OFFS3, 1, PORT_BTN2_HEIGHT * 2},
	{3, PORT_OFFS2, 1, PORT_BTN2_HEIGHT},
	{2, PORT_OFFS2, 1, PORT_BTN2_HEIGHT},
	{1, PORT_OFFS2, 1, PORT_BTN2_HEIGHT}
};

static const widget_pos_t landscape_wpos[] = {
	{1, LAND_OFFS0, 4, LAND_DISP_HEIGHT},
	{5, LAND_OFFS3, 2, LAND_BTN_HEIGHT},
	{0, LAND_OFFS1, 1, LAND_BTN_HEIGHT},
	{5, LAND_OFFS0, 2, LAND_DISP_HEIGHT},
	{0, LAND_OFFS0, 1, LAND_DISP_HEIGHT},
	{0, LAND_OFFS2, 1, LAND_BTN_HEIGHT * 2},
	{1, LAND_OFFS3, 1, LAND_BTN_HEIGHT},
	{2, LAND_OFFS3, 1, LAND_BTN_HEIGHT},
	{3, LAND_OFFS3, 1, LAND_BTN_HEIGHT},
	{1, LAND_OFFS2, 1, LAND_BTN_HEIGHT},
	{2, LAND_OFFS2, 1, LAND_BTN_HEIGHT},
	{3, LAND_OFFS2, 1, LAND_BTN_HEIGHT},
	{1, LAND_OFFS1, 1, LAND_BTN_HEIGHT},
	{2, LAND_OFFS1, 1, LAND_BTN_HEIGHT},
	{3, LAND_OFFS1, 1, LAND_BTN_HEIGHT},
	{4, LAND_OFFS3, 1, LAND_BTN_HEIGHT},
	{4, LAND_OFFS1, 1, LAND_BTN_HEIGHT * 2},
	{5, LAND_OFFS1, 1, LAND_BTN_HEIGHT * 2},
	{6, LAND_OFFS2, 1, LAND_BTN_HEIGHT},
	{6, LAND_OFFS1, 1, LAND_BTN_HEIGHT}
};

static void _delete_win_request_cb(void *data, Evas_Object *obj, void *event_info);
static void _navi_back_cb(void *data, Evas_Object *obj, void *event_info);
static void _popup_close_cb(void *data, Evas_Object *obj, void *event_info);
static void _keypad_click_cb(void *data, Evas_Object *obj, void *event_info);
static void _get_app_resource(const char *edj_file_in, char *edj_path_out);
static bool _create_main_layout(void);
static void _rearrange_main_layout(int orientation);
static bool _create_keypad(void);
static bool _create_display(void);
static void _display_popup(const char *text);

/**
 * @brief Sets the callback view's functions.
 * @param[in] get_display_string_func The callback function invoked when
 * the calculator's output needs to be displayed.
 * @param[in] key_pressed_func The callback function invoked when any key from
 * the keypad is pressed.
 */
void view_set_callbacks(get_display_string_cb get_display_string_func, key_pressed_cb key_pressed_func)
{
	s_info.get_display_string_func = get_display_string_func;
	s_info.key_pressed_func = key_pressed_func;
}

/**
 * @brief Creates essential objects: window, conformant and layout.
 */
Eina_Bool view_create(void *user_data)
{
	/* Create the window */
	s_info.win = view_create_win(PACKAGE);
	if (s_info.win == NULL) {
		dlog_print(DLOG_ERROR, LOG_TAG, "failed to create a window.");
		return EINA_FALSE;
	}

	/* Create the conformant */
	s_info.conform = view_create_conformant_without_indicator(s_info.win);
	if (s_info.conform == NULL) {
		dlog_print(DLOG_ERROR, LOG_TAG, "failed to create a conformant");
		return EINA_FALSE;
	}

	s_info.navi = view_create_naviframe(s_info.conform, "elm.swallow.content");
	if (!s_info.navi)
		return EINA_FALSE;

	if (!_create_main_layout())
		return EINA_FALSE;

	_rearrange_main_layout(elm_win_rotation_get(s_info.win));

	if (!_create_keypad())
		return EINA_FALSE;

	if (!_create_display())
		return EINA_FALSE;

	/* Show the window after main view is set up */
	evas_object_show(s_info.win);

	return EINA_TRUE;
}

/**
 * @brief Creates a basic window named package.
 * @param[in] pkg_name Name of the window
 */
Evas_Object *view_create_win(const char *pkg_name)
{
	int rots[4] = { 0, 90, 180, 270 };
	Evas_Object *win = NULL;

	/*
	 * Window
	 * Create and initialize elm_win.
	 * elm_win is mandatory to manipulate the window.
	 */
	win = elm_win_util_standard_add(pkg_name, pkg_name);
	elm_win_conformant_set(win, EINA_TRUE);
	elm_win_autodel_set(win, EINA_TRUE);

	/* Rotation setting */
	if (elm_win_wm_rotation_supported_get(win))
		elm_win_wm_rotation_available_rotations_set(win, (const int *)(&rots), 4);

	evas_object_smart_callback_add(win, "delete,request", _delete_win_request_cb, NULL);

	return win;
}

/**
 * @brief Creates a layout object for parent object based on provided EDJE script.
 * @param[in] parent The parent object for layout object.
 * @param[in] edj_file_name The relative path to the layout EDJE script.
 * @param[in] edj_group The name of the group to be loaded from the EDJE script.
 * @param[in] part_name The name of the EDJE part where the layout object has to be set.
 * @return The function returns layout object if it was created successfully,
 * otherwise 'NULL' is returned.
 */
Evas_Object *view_create_layout(Evas_Object *parent, const char *edj_file_name, const char *edj_group, const char *part_name)
{
	char edj_path[PATH_MAX] = {0, };
	Evas_Object *layout = NULL;

	_get_app_resource(edj_file_name, edj_path);

	layout = elm_layout_add(parent);
	if (!layout) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function elm_layout_add() failed.");
		return NULL;
	}

	if (!elm_layout_file_set(layout, edj_path, edj_group)) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function elm_layout_file_set() failed.");
		return NULL;
	}

	evas_object_size_hint_weight_set(layout, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);

	if (!part_name)
		elm_object_content_set(parent, layout);
	else
		elm_object_part_content_set(parent, part_name, layout);

	evas_object_show(layout);

	return layout;
}

/**
 * @brief Creates a conformant without indicator.
 * @param[in] win The object to which you want to set this conformant
 * Conformant is mandatory for base GUI to have proper size
 */
Evas_Object *view_create_conformant_without_indicator(Evas_Object *win)
{
	/*
	 * Conformant
	 * Create and initialize elm_conformant.
	 * elm_conformant is mandatory for base GUI to have proper size
	 * when indicator or virtual keypad is visible.
	 */
	Evas_Object *conform = NULL;

	if (win == NULL) {
		dlog_print(DLOG_ERROR, LOG_TAG, "window is NULL.");
		return NULL;
	}

	conform = elm_conformant_add(win);
	evas_object_size_hint_weight_set(conform, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
	elm_win_resize_object_add(win, conform);

	evas_object_show(conform);

	return conform;
}

/**
 * @brief Creates a naviframe for parent object and sets it to designated part.
 * @param[in] parent The parent object for naviframe.
 * @param[in] part_name The name of the part within parent object
 * where the naviframe is to be set,
 * @return This function returns a naviframe object if is was successfully created,
 * otherwise NULL is returned.
 */
Evas_Object *view_create_naviframe(Evas_Object *parent, const char *part_name)
{
	Evas_Object *navi = elm_naviframe_add(parent);
	if (!navi) {
		dlog_print(DLOG_ERROR, LOG_TAG, "elm_naviframe_add() failed.");
		return NULL;
	}

	elm_naviframe_prev_btn_auto_pushed_set(navi, EINA_FALSE);

	eext_object_event_callback_add(navi, EEXT_CALLBACK_BACK, _navi_back_cb, NULL);
	elm_object_part_content_set(parent, part_name, navi);

	return navi;
}

/**
 * @brief Creates a background for parent object.
 * @param[in] parent The parent object for the background.
 * @return This function returns a background object if is was successfully created,
 * otherwise NULL is returned.
 */
Evas_Object *view_create_background(Evas_Object *parent)
{
	Evas_Object *bg = elm_bg_add(parent);
	if (!bg) {
		dlog_print(DLOG_ERROR, LOG_TAG, "elm_bg_add() failed.");
		return NULL;
	}

	evas_object_size_hint_align_set(bg, EVAS_HINT_FILL, EVAS_HINT_FILL);
	evas_object_show(bg);

	return bg;
}

/**
 * @brief Creates a table for parent object or its item and sets it to designated part.
 * @param[in] parent_widget The parent object for the table.
 * @param[in] parent_item The parent object's item for the table.
 * @param[in] part_name The name of the part within parent object or its item
 * where the table is to be set,
 * @return This function returns a table object if is was successfully created,
 * otherwise NULL is returned.
 */
Evas_Object *view_create_table(Evas_Object *parent_widget, Elm_Object_Item *parent_item, const char *part_name)
{
	if (parent_item)
		parent_widget = elm_object_item_widget_get(parent_item);

	Evas_Object *table = elm_table_add(parent_widget);
	if (!table) {
		dlog_print(DLOG_ERROR, LOG_TAG, "elm_table_add() failed.");
		return NULL;
	}

	elm_table_homogeneous_set(table, EINA_TRUE);
	evas_object_size_hint_weight_set(table, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);

	if (part_name) {
		if (parent_item)
			elm_object_item_part_content_set(parent_item, part_name, table);
		else if (parent_widget)
			elm_object_part_content_set(parent_widget, part_name, table);
	}

	evas_object_show(table);

	return table;
}

/**
 * @brief Creates a grid for parent object.
 * @param[in] parent The parent object for the grid.
 * @return This function returns a grid object if is was successfully created,
 * otherwise NULL is returned.
 */
Evas_Object *view_create_grid(Elm_Object_Item *parent)
{
	Evas_Object *grid = elm_grid_add(parent);
	if (!grid) {
		dlog_print(DLOG_ERROR, LOG_TAG, "elm_grid_add() failed.");
		return NULL;
	}

	evas_object_size_hint_align_set(grid, EVAS_HINT_FILL, EVAS_HINT_FILL);
	evas_object_size_hint_padding_set(grid, VIEW_BORDER_WIDTH, VIEW_BORDER_WIDTH, VIEW_BORDER_WIDTH, VIEW_BORDER_WIDTH);
	evas_object_show(grid);

	return grid;
}

/**
 * @brief Creates a button object for the parent object.
 * @param[in] parent The parent object for button object.
 * @param[in] caption The caption to be displayed on the button.
 * @param[in] on_click_cb The callback function's handler to be invoked on button click.
 * @param[in] data The data to be passed to the callback function.
 * @return The function returns button object if it was created successfully,
 * otherwise 'NULL' is returned.
 */
Evas_Object *view_create_button(Evas_Object *parent, const char *caption, Evas_Smart_Cb on_click_cb, void *data)
{
	Evas_Object *button = elm_button_add(parent);
	if (!button) {
		dlog_print(DLOG_ERROR, LOG_TAG, "elm_button_add() failed.");
		return NULL;
	}

	evas_object_size_hint_align_set(button, EVAS_HINT_FILL, EVAS_HINT_FILL);
	evas_object_size_hint_padding_set(button, WIDGET_PADDING, WIDGET_PADDING, WIDGET_PADDING, WIDGET_PADDING);
	elm_object_text_set(button, caption);
	evas_object_smart_callback_add(button, "clicked", on_click_cb, data);
	evas_object_show(button);

	return button;
}

/**
 * @brief Creates a label object for the parent object.
 * @param[in] parent The parent object for label object.
 * @param[in] caption The caption to be displayed on the label.
 * @return The function returns label object if it was created successfully,
 * otherwise 'NULL' is returned.
 */
Evas_Object *view_create_label(Evas_Object *parent, const char *caption)
{
	Evas_Object *label = elm_label_add(parent);
	if (!label) {
		dlog_print(DLOG_ERROR, LOG_TAG, "elm_label_add() failed.");
		return NULL;
	}

	if (caption)
		elm_object_text_set(s_info.display, caption);

	evas_object_size_hint_align_set(label, 1, 0.5);
	evas_object_size_hint_padding_set(label, WIDGET_PADDING, WIDGET_PADDING + DISP_NUMBER_RIGHT_PADDING, WIDGET_PADDING, WIDGET_PADDING);
	evas_object_show(label);

	return label;
}

/**
 * @brief Function which prints the calculator's output on the display.
 */
void view_update_display(void)
{
	char buff[DISP_STR_SIZE] = {'\0'};
	const char *calc_output = NULL;

	if (!s_info.get_display_string_func) {
		dlog_print(DLOG_ERROR, LOG_TAG, "get_display_string_func callback function is not attached.");
		return;
	}

	calc_output = s_info.get_display_string_func();

	snprintf(buff, DISP_STR_SIZE, "<font_size=%d>%s</font_size>", (int)DISP_NUMBER_TEXT_SIZE, calc_output);
	elm_object_text_set(s_info.display, buff);
}

/**
 * @brief Function updates the text strings according to regional formatting rules.
 */
void view_change_region_format(const char *format)
{
	elm_object_text_set(s_info.button_decimal, format);
}

/**
 * @brief Rearranges the application's layout according to the orientation provided.
 */
void view_change_orientation(int orientation)
{
	_rearrange_main_layout(orientation);
}

/**
 * @brief Destroys window and frees its resources.
 */
void view_destroy(void)
{
	if (s_info.win == NULL)
		return;

	evas_object_del(s_info.win);
}

/**
 * @brief Internal callback function invoked when the main window needs to be destroyed.
 * @param[in] data The user data passed to the evas_object_smart_callback_add() function.
 * @param[in] obj The object invoking this callback function.
 * @param[in] event_info The structure containing the information on this event.
 */
static void _delete_win_request_cb(void *data, Evas_Object *obj, void *event_info)
{
	ui_app_exit();
}

/**
 * @brief Internal callback function invoked when the HW Back button is pressed.
 * @param[in] data The user data passed to the eext_object_event_callback_add() function.
 * @param[in] obj The object invoking this callback function.
 * @param[in] event_info The structure containing the information on this event.
 */
static void _navi_back_cb(void *data, Evas_Object *obj, void *event_info)
{
	if (elm_naviframe_top_item_get(obj) == elm_naviframe_bottom_item_get(obj))
		elm_win_lower(s_info.win);
	else
		elm_naviframe_item_pop(obj);
}

/**
 * @brief Internal callback function invoked when the HW Back button
 * or 'OK' button is pressed while popup window is visible.
 * @param[in] data The user data passed to the callback attachment function.
 * @param[in] obj The object invoking this callback function.
 * @param[in] event_info The structure containing the information on this event.
 */
static void _popup_close_cb(void *data, Evas_Object *obj, void *event_info)
{
	evas_object_del((Evas_Object *)data);
}

/**
 * @brief Internal callback function invoked when calculator keypad button is pressed.
 * Internally, this function prints the output to the display.
 * @param[in] data The user data passed to the evas_object_smart_callback_add() function.
 * @param[in] obj The object invoking this callback function.
 * @param[in] event_info The structure containing the information on this event.
 */
static void _keypad_click_cb(void *data, Evas_Object *obj, void *event_info)
{
	key_type key = (key_type)data;
	char msg[DISP_STR_SIZE] = {'\0'};

	if (!s_info.key_pressed_func) {
		dlog_print(DLOG_ERROR, LOG_TAG, "key_pressed_func callback function is not attached.");
		return;
	}

	if (!s_info.key_pressed_func(key)) {
		snprintf(msg, DISP_STR_SIZE, "<p align=center>Can not input more<br/>than %d digits</p>", DISP_MAX_DIGITS);
		_display_popup(msg);
	}
}

/**
 * @brief Internal function which creates fully qualified path to the provided resource file.
 * @param[in] edj_file_in The file name.
 * @param[out] edj_path_out The fully qualified path to the edj_file_in file.
 */
static void _get_app_resource(const char *edj_file_in, char *edj_path_out)
{
	char *res_path = app_get_resource_path();
	if (res_path) {
		snprintf(edj_path_out, PATH_MAX, "%s%s", res_path, edj_file_in);
		free(res_path);
	}
}

/**
 * @brief Internal function which creates the main layout based on standard containers
 * composition. The combination of tables and grid is used, where the table plays a role
 * of the main container for the grid which stores an additional table within each cell.
 * Each table has only one cell for single object storage. This kind of objects hierarchy
 * enables padding addition to cell content.
 * @return This function returns 'true' if the main layout was created successfully,
 * otherwise 'false' is returned.
 */
static bool _create_main_layout(void)
{
	int i;
	Evas_Object *bg;
	Evas_Object *table;
	Elm_Object_Item *item = elm_naviframe_item_push(s_info.navi, "Calculator", NULL, NULL, NULL, NULL);

	/* Create single cell table so the internal grid will support padding. */
	table = view_create_table(NULL, item, "default");
	if (!table)
		return EINA_FALSE;

	/* Add background to the view with default theme color. */
	bg = view_create_background(table);
	if (!bg)
		return EINA_FALSE;

	elm_table_pack(table, bg, 0, 0, 1, 1);

	/* Create grid for holding and aligning the widgets. */
	s_info.grid = view_create_grid(table);
	if (!s_info.grid)
		return EINA_FALSE;

	elm_table_pack(table, s_info.grid, 0, 0, 1, 1);

	for (i = 0; i < WIDGETS_COUNT; i++) {
		s_info.keypad[i] = view_create_table(s_info.grid, NULL, NULL);
		if (!s_info.keypad[i])
			return false;
	}

	return true;
}

/**
 * @brief Internal function which changes the sizes and location of the keypad
 * widgets containers (tables) in order to accommodate the view to the current
 * device orientation.
 * @param[in] orientation The device orientation in degrees unit.
 */
static void _rearrange_main_layout(int orientation)
{
	int i;
	const widget_pos_t *pos_array = NULL;

	/* Remove all widget from the grid without deleting. */
	elm_grid_clear(s_info.grid, EINA_FALSE);

	/* Obtaining widget position parameters and resizing the grid
	 * according to the device orientation. */
	if ((((abs(orientation) + CIRCLE_HALF_QUARTER_DEG) / CIRCLE_QUARTER_DEG) & 1) == 0) {
		pos_array = portrait_wpos;
		elm_grid_size_set(s_info.grid, PORT_GRID_W, PORT_GRID_H);
	} else {
		pos_array = landscape_wpos;
		elm_grid_size_set(s_info.grid, LAND_GRID_W, LAND_GRID_H);
	}

	/* Packing widgets to the grid using information from the array */
	for (i = 0; i < WIDGETS_COUNT; i++)
		elm_grid_pack(s_info.grid, s_info.keypad[i], pos_array[i].x, pos_array[i].y, pos_array[i].w, pos_array[i].h);
}

/**
 * @brief Internal function which creates the keypad.
 * For each layout's container created with the _create_main_layout() function,
 * an appropriate widget (button) is created and packed into that
 * container.
 * @return This function returns 'true' if the keypad was created successfully,
 * otherwise 'false' is returned.
 */
static bool _create_keypad(void)
{
	int i;
	const char *caption;
	Evas_Object *button;

	for (i = 0; i < WIDGETS_COUNT; i++)
		if (widget_params[i].key != KEY_NONE) {
			caption = widget_params[i].text;

			if (widget_params[i].key == KEY_DEC)
				caption = ".";

			button = view_create_button(s_info.keypad[i], caption, _keypad_click_cb, (void *)widget_params[i].key);
			if (!button)
				return false;

			elm_table_pack(s_info.keypad[i], button, 0, 0, 1, 1);

			if (widget_params[i].key == KEY_DEC)
				s_info.button_decimal = button;
		}

	return true;
}

/**
 * @brief Internal function which creates the calculator's display.
 * The created display is packed into the container previously created
 * with _create_main_layout() function, whose key identifier does not point
 * to any keypad key.
 * @return This function returns 'true' if the display was created successfully,
 * otherwise 'false' is returned.
 */
static bool _create_display(void)
{
	int i;

	for (i = 0; i < WIDGETS_COUNT; i++)
		if (widget_params[i].key == KEY_NONE) {
			s_info.display = view_create_label(s_info.keypad[i], NULL);
			if (!s_info.display)
				return false;

			elm_table_pack(s_info.keypad[i], s_info.display, 0, 0, 1, 1);

			view_update_display();
			return true;
		}

	return false;
}

/**
 * @brief Internal function which displays the popup window with provided text message
 * and one button for its closure.
 * @param[in] text The popup window's text caption.
 */
static void _display_popup(const char *text)
{
	Evas_Object *button;
	Evas_Object *popup = elm_popup_add(s_info.win);
	if (!popup) {
		dlog_print(DLOG_ERROR, LOG_TAG, "elm_popup_add() failed.");
		return;
	}

	elm_popup_align_set(popup, ELM_NOTIFY_ALIGN_FILL, 1.0);

	elm_object_part_text_set(popup, "default", text);

	button = elm_button_add(popup);
	elm_object_text_set(button, "OK");
	elm_object_part_content_set(popup, "button1", button);

	evas_object_show(button);
	evas_object_show(popup);

	evas_object_smart_callback_add(button, "clicked", _popup_close_cb, popup);
	eext_object_event_callback_add(popup, EEXT_CALLBACK_BACK, _popup_close_cb, popup);
}
